
package require tdom
package require htmlparse

namespace eval html2tdom {
    variable xsltcmd
    variable topdir [file dirname [info script]]
}

proc html2tdom::parse { html } {
    variable xsltcmd
    variable topdir

    set baseurl ""
    regexp {(?in)^\s*SourceURL:\s*(.+)$} [string range $html 0 200] {} baseurl
    regsub -all {<!\[if\s+?(!supportLists|gte vml 1)\].*?<!\[endif\]>} $html {} html
    regsub -all {<!\[if\s+?[^\]]*?\]>} $html {} html
    regsub -all {<!\[endif\]>} $html {} html
    regsub -all {<!--StartFragment-->} $html {<StartFragment></StartFragment>} html
    regsub -all {<!--EndFragment-->} $html {<EndFragment></EndFragment>} html

    catch {html2tdom::tags destroy}

    ::struct::stack html2tdom::tags

    set doc [dom createDocument html]
    html2tdom::tags push [$doc documentElement]
    htmlparse::parse -cmd [list html2tdom::2tdomCallback $doc] $html
    Reorder [$doc documentElement]
    html2tdom::tags destroy

    if { ![info exists xsltcmd] } {
	set filename [file join $topdir html2docbook.xslt]
	set xsltdoc [dom parse [tDOM::xmlReadFile $filename]]
	set xsltcmd [$xsltdoc toXSLTcmd]
    }

    foreach node [$doc selectNodes {//*[@style]}] {
	set font-family [_give_syle_item [$node @style] font-family]
	if { ${font-family} eq "symbol" } {
	    foreach descendantNode [$node selectNodes .//text()] {
		package require xml2pdf
		set t [xml2pdf::symbol2unicode [$descendantNode nodeValue]]
		$descendantNode nodeValue $t
	    }
	}
    }
    set newdoc [$xsltcmd $doc]
    $doc delete

    # some clean up

    set xp {//listitem/*[last() and name()='para']}
    foreach node [$newdoc selectNodes $xp] {
	$node delete
    }

    return $newdoc
}

proc html2tdom::_give_syle_item { style item } {
    foreach i [split $style ";"] {
	foreach "n v" [split $i :] {
	    set n [string trim [string tolower $n]]
	    set v [string trim [string tolower $v]]
	    if { $item eq $n } {
		return $v
	    }
	}
    }
    return ""
}

proc html2tdom::2tdomCallback { doc tag slash param textBehindTheTag } {

    # Normalize tag information for later comparisons. Also remove
    # superfluous whitespace. Don't forget to decode the standard
    # entities.

    set tag [string tolower $tag]
    if { $tag eq "hmstart" } { return }
    if {$textBehindTheTag ne "" } {
	set textBehindTheTag [htmlparse::mapEscapes $textBehindTheTag]
	regsub -all {\s*\n\s*} $textBehindTheTag { } textBehindTheTag

    }

    if { $slash eq "/"} {
	# Handle closing tags. Standard operation is to pop the tag
	# from the stack of open tags. We don't do this for </p> and
	# </li>. As they were optional they were never pushed onto the
	# stack (Well, actually they are just popped immediately after
	# they were pusheed, see below).

	switch -exact -- $tag {
	    base - option - meta - li - p {
		# Ignore, nothing to do.                
	    }
	    default {
		# The moment we get a closing tag which does not match
		# the tag on the stack we have two possibilities on how
		# this came into existence to choose from:
		#
		# a) A tag is now closed but was never opened.
		# b) A tag requiring an end tag was opened but the end
		#    tag was omitted and we now are at a tag which was
		#    opened before the one with the omitted end tag.

		# NOTE:
		# Pages delivered from the amazon.uk site contain both
		# cases: </a> without opening, <b> & <font> without
		# closing. Another error: <a><b></a></b>, i.e. overlapping
		# tags. Fortunately this can be handled by the algorithm
		# below, in two cycles, one of which is case (b), followed
		# by case (a). It seems as if Amazon/UK believes that visual
		# markup like <b> and <font> is an option (switch-on) instead
		# of a region.

		# Algorithm used here to deal with these:
		# 1) Search whole stack for the matching opening tag.
		#    If there is one assume case (b) and pop everything
		#    until and including this opening tag.
		# 2) If no matching opening tag was found assume case
		#    (a) and ignore the tag.
		#
		# Part (1) also subsumes the normal case, i.e. the
		# matching tag is at the top of the stack.

		set nodes [html2tdom::tags peek [html2tdom::tags size]]
		# Note: First item is top of stack, last item is bottom of stack !
		# (This behaviour of tcllib stacks is not documented
		# -> we should update the manpage).

		#foreach n $nodes {lappend tstring [p get $n -key type]}
		#puts stderr --[join $tstring]--

		set level 1
		set found 0
		foreach n $nodes {
		    set name [$n nodeName]
		    if {0 == [string compare $tag $name]} {
		        # Found an earlier open tag -> (b).
		        set found 1
		        break
		    }
		    incr level
		}
		if {$found} {
		    html2tdom::tags pop $level
		    if {$level > 1} {
		        #foreach n $nodes {lappend tstring [$tree get $n type]}
		        #puts stderr "\tdesync at <$tag> ($tstring) => pop $level"
		    }
		} else {
		    #foreach n $nodes {lappend tstring [$tree get $n type]}
		    #puts stderr "\tdesync at <$tag> ($tstring) => ignore"
		}
	    }
	}

	# If there is text behind a closing tag X it belongs to the
	# parent tag of X.

	if {$textBehindTheTag ne "" } {
	    # Attach the text behind the closing tag to the reopened
	    # context.

	    set parentNode [html2tdom::tags peek]
	    set newTextNode [$doc createTextNode $textBehindTheTag]
	    $parentNode appendChild $newTextNode
	}

    } else {
	# Handle opening tags. The standard operation for most is to
	# push them onto the stack and thus open a nested context.
	# This does not happen for both the optional tags (p, li) and
	# the ones which don't have closing tags (meta, br, option,
	# input, area, img).
	#
	# The text coming with the tag will be added after the tag if
	# it is a tag without a matching close, else it will be added
	# as a node below the tag (as it is the region between the
	# opening and closing tag and thus nested inside). Empty text
	# is ignored under all circcumstances.

	set parentNode [html2tdom::tags peek]

	set tagsList [list td tr]
	while 1 {
	    set pos1 [lsearch -exact $tagsList $tag]
	    set pos2 [lsearch -exact $tagsList [$parentNode nodeName]]
	    if { $pos1 == -1 || $pos2 == -1 || $pos1 < $pos2 } { break }
	    html2tdom::tags pop
	    set parentNode [html2tdom::tags peek]
	}

	set rex {(\w+)\s*=\s*([^\s\"]+|\"[^\"]*\"|\'[^\']*\')}
	foreach "- name value" [regexp -inline -all $rex $param] {
	    set atts($name) [string trim [htmlparse::mapEscapes $value] '\"]
	}
	set rex {mso-list:\S+\s+level(\d+)}
	if { $tag eq "html" } {
	    set newNode $parentNode
	    if { [array size atts] } { eval [list $newNode setAttribute] [array get atts] }
	} elseif { [info exists atts(style)] && [regexp $rex $atts(style) {} level] } {
	    set newNode $parentNode
	    for { set i 1 } { $i < $level } { incr i } {
		$newNode appendChild [set n [$doc createElement blockquote]]
		set newNode $n
	    }
	    $newNode appendChild [set n [$doc createElement ul]]
	    set newNode $n
	    if { [array size atts] } { eval [list $newNode setAttribute] [array get atts] }
	    $newNode appendChild [set n [$doc createElement li]]
	    set newNode $n
	} elseif { [regexp {^\?} $tag] } {
	    set newNode [$doc createElement span]
	    $parentNode appendChild $newNode
	} else {
	    set newNode [$doc createElement $tag]
	    if { [array size atts] } { eval [list $newNode setAttribute] [array get atts] }
	    $parentNode appendChild $newNode
	}

	if {$textBehindTheTag ne "" } {
	    switch -exact -- $tag {
		input - area - img - br {
		    set parent $parentNode
		}
		default {
		    set parent $newNode
		}
	    }
	    set newTextNode [$doc createTextNode $textBehindTheTag]
	    $parent appendChild $newTextNode
	}
	html2tdom::tags push $newNode

	# Special handling: <p>, <li> may have no closing tag => pop
	#                 : them immediately.
	#
	# Special handling: <meta>, <br>, <option>, <input>, <area>,
	#                 : <img>: no closing tags for these.

	switch -exact -- $tag {
	    hr - base - meta - li - br - option - input - area - img - p - h1 - h2 - h3 - h4 - h5 - h6 {
		html2tdom::tags pop
	    }
	    default {}
	}
    }
}

proc html2tdom::Reorder { node } {
    set tags [list h1 h2 h3 h4 h5 h6 p li]
    switch -exact -- [$node nodeName] {
	h1 - h2 - h3 - h4 - h5 - h6 - p - li {
	    # Look for right siblings until the next node with the
	    # same type (or end of level) and move these below this
	    # node.

	    while {1} {
		set sibling [$node nextSibling]
		if { $sibling eq "" || [lsearch -exact $tags [$sibling nodeName]] != -1 } {
		    break
		}
		$node appendChild $sibling
	    }
	}
	default {
	    # Ignore tag
	}
    }
    foreach childNode [$node childNodes] {
	Reorder $childNode
    }
}

namespace eval docbook2txt {
    variable xsltcmd
    variable topdir [file dirname [info script]]
}

proc docbook2txt::go { xml } {
    variable xsltcmd
    variable topdir

    set doc [dom parse $xml]

    if { ![info exists xsltcmd] } {
	set filename [file join $topdir docbook2txt.xslt]
	set xsltdoc [dom parse [tDOM::xmlReadFile $filename]]
	set xsltcmd [$xsltdoc toXSLTcmd]
    }
    set newdoc [$xsltcmd $doc]
    $doc delete
    set txt [$newdoc asText]

    # this is a dirty trick because tdom XSLT does not work
    # well when there is only space inside <xsl:text>
    regsub -all -- {---NEWLINE---} $txt "\n" txt
    $newdoc delete
    return $txt
}























